/*
 *	Copyright (C) 2020 3D Repo Ltd
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Affero General Public License as
 *	published by the Free Software Foundation, either version 3 of the
 *	License, or (at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Affero General Public License for more details.
 *
 *	You should have received a copy of the GNU Affero General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "RepoJobs.h"
#include "Repo3d.h"
#include "RepoTypes.h"
#include "RepoWebRequestManager.h"

RepoJob::RepoJob(FName Name)
{
	IsCompleted = false;
	Progress = 0;
	TotalJobs = 1;
	CompletedJobs = 0;
	OverallProgress = 0;
	this->Name = Name;

	UE_LOG(Log3DRepo, VeryVerbose, TEXT("Start %s"), *Name.ToString())
}

RepoJob::~RepoJob()
{
}

void RepoJob::Update(float progress)
{
	Progress = progress;
	Update();
}

void RepoJob::Complete()
{
	IsCompleted = true;
	Progress = 1;
	Update();
}

void RepoJob::Add(RepoJobRef SubJob)
{
	if (Children.Contains(SubJob))
	{
		UE_LOG(Log3DRepo, Warning, TEXT("Attemping to add Job %s twice to Job %s. This is likely a bug."), *this->Name.ToString(), *SubJob->Name.ToString());
	}
	if (IsCompleted) 
	{
		UE_LOG(Log3DRepo, Error, TEXT("Attempting to add Job %s to completed Job %s. Child jobs may not be added to Jobs marked as Complete."), *this->Name.ToString(), *SubJob->Name.ToString());
	}
	Children.Add(SubJob);
	SubJob->Parents.Add(This);
	Update();
}

void RepoJob::Update()
{
	TotalJobs = 1; // 1 is for this Job.
	
	CompletedJobs = 0;
	if (IsCompleted)
	{
		CompletedJobs++;
	}
	
	OverallProgress = 0;
	float OverallProgressSamples = 0;

	for (auto Child : Children)
	{
		TotalJobs += Child->TotalJobs;	
		OverallProgress += Child->OverallProgress;
		OverallProgressSamples++;
		
		if (Child->IsCompleted)
		{
			CompletedJobs += Child->CompletedJobs;
		}
	}

	OverallProgress += FMath::Clamp(Progress, 0.0f, 1.0f);
	OverallProgressSamples++;

	OverallProgress /= OverallProgressSamples;

	bool HierarchyComplete = IsComplete();

	if (HierarchyComplete)
	{
		UE_LOG(Log3DRepo, VeryVerbose, TEXT("Completed %s"), *Name.ToString())
	}

	if (OnProgressUpdate.IsBound())
	{
		OnProgressUpdate.Broadcast(This.Pin());
		if (HierarchyComplete)
		{
			// In case the same job has been added multiple times at the same level, stop firing once
			// the update for the final job completion is set.
			OnProgressUpdate.Clear();
		}
	}

	for (auto ParentPtr : Parents)
	{
		auto Parent = ParentPtr.Pin();
		if (Parent.IsValid()) {
			Parent->Update();
		}
	}
}

bool RepoJob::IsComplete()
{
	if (IsCompleted)
	{
		for (auto Child : Children)
		{
			if (!Child->IsComplete())
			{
				return false;
			}
		}
		return true;
	}
	return false;
}

RepoJobRef RepoJob::Create(FName JobName)
{
	auto Job = MakeShared<RepoJob, ESPMode::ThreadSafe>(JobName);
	Job->This = TWeakPtr<RepoJob, ESPMode::ThreadSafe>(Job);
	return Job;
}

RepoJobRef RepoJob::Create(FString JobName)
{
	return Create(FName(*JobName));
}

RepoJobRef RepoJob::CreateCompleted(FString JobName)
{
	auto Job = Create(JobName);
	Job->Complete();
	return Job;
}

RepoDownloadJobRef RepoDownloadJob::Create(FString Name)
{
	auto Job = MakeShared<RepoDownloadJob, ESPMode::ThreadSafe>(FName(*Name));
	Job->This = TWeakPtr<RepoJob, ESPMode::ThreadSafe>(Job);
	Job->TotalBytesRecieved = 0;
	return Job;
}

RepoDownloadJob::RepoDownloadJob(FName Name) :
	RepoJob(Name)
{
}

void RepoDownloadJob::Update(int32 ContentLength, int32 BytesRecieved)
{
	RepoJob::Update((float)BytesRecieved / (float)ContentLength);
}

void RepoDownloadJob::Complete(const FString& Uri, int32 ResponseCode)
{
	RepoJob::Complete();
	if (ResponseCode == 200)
	{
		UE_LOG(Log3DRepo, Verbose, TEXT("Fetched %s (%d)"), *Uri, ResponseCode);
	}
	else
	{
		UE_LOG(Log3DRepo, Error, TEXT("Failed to fetch %s (%d)"), *Uri, ResponseCode);
	}
}

RepoRequestJob::RepoRequestJob(FString Uri) :
	RepoJob(FName(*(TEXT("Download & Handle ") + Uri))),
	DownloadJob(RepoDownloadJob::Create(TEXT("Download ") + Uri)),
	CallbackJob(RepoJob::Create(TEXT("Handle ") + Uri))
{
}

RepoRequestJobRef RepoRequestJob::Create(FString Name)
{
	auto Job = MakeShared<RepoRequestJob, ESPMode::ThreadSafe>(Name);
	Job->This = TWeakPtr<RepoJob, ESPMode::ThreadSafe>(Job);
	Job->Add(Job->DownloadJob);
	Job->Add(Job->CallbackJob);
	Job->Complete();
	return Job;
}

void RepoRequestJob::SetUri(FString Uri)
{
	UE_LOG(Log3DRepo, Verbose, TEXT("Fetching %s"), *Uri);
}


RepoLoadModelJob::RepoLoadModelJob(FLoadModel& Params)
	:RepoJob(FName(*(TEXT("Load Model ") + Params.Model))), Params(Params)
{
}

RepoLoadModelJobRef RepoLoadModelJob::Create(FLoadModel& Params)
{
	auto Job = MakeShared<RepoLoadModelJob, ESPMode::ThreadSafe>(Params);
	Job->This = TWeakPtr<RepoJob, ESPMode::ThreadSafe>(Job);
	return Job;
}