/*
 *	Copyright (C) 2020 3D Repo Ltd
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Affero General Public License as
 *	published by the Free Software Foundation, either version 3 of the
 *	License, or (at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Affero General Public License for more details.
 *
 *	You should have received a copy of the GNU Affero General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "RepoSrcImporter.h"
#include "Misc/Compression.h"
#include "HAL/UnrealMemory.h"
#include "Materials/MaterialInstanceDynamic.h"
#include <string>
#include "Misc/CString.h"


DECLARE_CYCLE_STAT(TEXT("Handle SRC"), STAT_HandleSRC, STATGROUP_Repo3D);
DECLARE_CYCLE_STAT(TEXT("Generate Mesh"), STAT_GenerateMesh, STATGROUP_Repo3D);
DECLARE_FLOAT_ACCUMULATOR_STAT(TEXT("Last Mappings Response Time (ms)"), STAT_DownloadMappings, STATGROUP_Repo3D);
DECLARE_FLOAT_ACCUMULATOR_STAT(TEXT("Last SRC Response Time (ms)"), STAT_DownloadSRC, STATGROUP_Repo3D);
DECLARE_DWORD_ACCUMULATOR_STAT(TEXT("Num Triangles"), STAT_TotalTriangles, STATGROUP_Repo3D);
DECLARE_DWORD_ACCUMULATOR_STAT(TEXT("Num Vertices"), STAT_TotalVertices, STATGROUP_Repo3D);
DECLARE_DWORD_ACCUMULATOR_STAT(TEXT("Num Objects"), STAT_TotalObjects, STATGROUP_Repo3D);
DECLARE_MEMORY_STAT(TEXT("Uncompressed"), STAT_Uncompressed, STATGROUP_Repo3D);

// UE5 has introduced 
#if ENGINE_MAJOR_VERSION < 5
typedef FVector FVector3f;
typedef FVector2D FVector2f;
#endif

#define SRC_MAGIC_BIT_COMPRESSED 24
#define SRC_MAGIC_BIT_UNCOMPRESSED 23
#define SRC_VERSION 42
#define SRC_COMPONENT_SCALAR 5126
#define SRC_PREAMBLE_BYTES 12

void URepoSrcImporter::BeginDestroy()
{
	Super::BeginDestroy();
	// URepoSrcImporter cleanup 
}

RepoJobRef URepoSrcImporter::RequestRevision(FString teamspace, FString model, FString revision)
{
	auto Job = RepoJob::Create(TEXT("Get Assets ") + teamspace + model + revision);

	check(manager.IsValid());

	// in parallel, set the model units
	Job->Add(RepoWebRequestHelpers::GetModelSettings(
		manager.ToSharedRef(), teamspace, model,
		RepoWebRequestHelpers::ModelSettingsDelegate::CreateUObject(this, &URepoSrcImporter::HandleModelSettings)
	));

	// Get the src assets file which is need to to figure out what other files
	// need to be downloaded, and where from
	if (!revision.IsEmpty())
	{
		Job->Add(manager->GetRequest(
			FString::Printf(TEXT("%s/%s/revision/%s/srcAssets.json"), *teamspace, *model, *revision),
			RepoWebRequestDelegate::CreateUObject(this, &URepoSrcImporter::AssetsRequestCompleted)
		));
	}
	else
	{
		Job->Add(manager->GetRequest(
			FString::Printf(TEXT("%s/%s/revision/master/head/srcAssets.json"), *teamspace, *model),
			RepoWebRequestDelegate::CreateUObject(this, &URepoSrcImporter::AssetsRequestCompleted)
		));
	}

	Job->Complete();
	return Job;
}

void URepoSrcImporter::HandleModelSettings(TSharedRef<RepoWebRequestHelpers::ModelSettings> Settings)
{
	// Unreal's world units are cm
	actor->Units = Settings->Units;
	double factor = FUnitConversion::Convert(1.f, Settings->Units, EUnit::Centimeters);
	actor->SetActorScale3D(FVector(factor, factor, factor));
}

RepoJobRef URepoSrcImporter::HandleAssets(const FString& string)
{
	TSharedPtr<FJsonObject> assetsList = MakeShareable(new FJsonObject());
	TSharedRef<TJsonReader<TCHAR>> reader = TJsonReaderFactory<TCHAR>::Create(string);
	FJsonSerializer::Deserialize(reader, assetsList);

	// Revision might be a federation, so there may be multiple models to load

	auto Job = RepoJob::Create(FName(TEXT("Import Assets")));
	for (auto model : assetsList->GetArrayField("models"))
	{
		FModelInfo modelInfo;
		modelInfo.ModelId = model->AsObject()->GetStringField("model");
		modelInfo.Teamspace = model->AsObject()->GetStringField("database");

		auto offsetElements = model->AsObject()->GetArrayField("offset");
		for (int32 i = 0; i < offsetElements.Num(); i++)
		{
			modelInfo.Offset[i] = offsetElements[i]->AsNumber();
		}

		actor->ModelIdToModelInfoMap.Add(modelInfo.ModelId, modelInfo);
		
		OnModel.ExecuteIfBound(modelInfo);

		// Kick off a RepoSrcAssetImporter to process the files relating to this asset
		for (auto asset : model->AsObject()->GetArrayField("assets"))
		{
			auto importer = MakeShared<RepoSrcAssetImporter>(manager, TextureCache);

			importer->SetActor(actor);
			importer->SetMaterialPrototype(materialOpaque, materialTranslucent);
			importer->SetCollisionProfileName(CollisionProfileName);
			importers.Add(importer);

			importer->OnComplete.BindUObject(this, &URepoSrcImporter::HandleCompleted, importer);

			importer->SetOffset(modelInfo.Offset); // this method automatically performs the coordinate transform from server coordinate system to Unreal
			importer->SetUri(
				modelInfo.Teamspace,
				modelInfo.ModelId,
				asset->AsString()
			);
			Job->Add(importer->RequestSrc());
		}
	}

	Job->Complete();
	return Job;
}

void URepoSrcImporter::HandleCompleted(TSharedRef<RepoSrcAssetImporter> importer)
{
	importers.Remove(importer);
	if(importers.Num() <= 0)
	{
		OnComplete.ExecuteIfBound();
	}
}

void URepoSrcImporter::AssetsRequestCompleted(RepoWebResponsePtr Result)
{
	if (Result->bWasSuccessful && Result->Response->GetResponseCode() == 200)
	{
		Result->Job->Add(
			HandleAssets(Result->Response->GetContentAsString())
		);
	}
	else
	{
		OnComplete.ExecuteIfBound(); // If we failed to get the Assets then abort. Flag that we are complete anyway though so at least the user knows there are no more attempts running.
	}
}

void RepoSrcAssetImporter::SetOffset(FVector v)
{
	Offset = TransformCoordinateSystem(v);
}

RepoJobRef RepoSrcAssetImporter::RequestSrc()
{
	// For this version, just read the mapping first, and only when its received request the main SRC file
	return manager->GetRequest(
		FString::Printf(TEXT("%s.json.mpc"), *Uri),
		RepoWebRequestDelegate::CreateLambda(
			[this](TSharedPtr<RepoWebResponse> result)
			{	
				if (MappingRequestCompleted(result)) {
					result->Job->Add(manager->GetRequest(
						FString::Printf(TEXT("%s.src.mpc"), *Uri),
						RepoWebRequestDelegate::CreateRaw(this, &RepoSrcAssetImporter::SrcRequestCompleted)
					));
				}
				else
				{
					OnComplete.ExecuteIfBound();
				}
			})
	);
}

void RepoSrcAssetImporter::SrcRequestCompleted(RepoWebResponsePtr Result)
{
	SET_FLOAT_STAT(STAT_DownloadSRC, Result->Time * 1000.0);
	
	if (Result->bWasSuccessful && Result->Response->GetResponseCode() == 200)
	{
		Result->Job->Add(HandleSrc(Result->Response->GetContent()));
	}

	OnComplete.ExecuteIfBound();
}

bool RepoSrcAssetImporter::MappingRequestCompleted(TSharedPtr<RepoWebResponse> Result)
{
	SET_FLOAT_STAT(STAT_DownloadMappings, Result->Time * 1000.0);

	if (Result->bWasSuccessful && Result->Response->GetResponseCode() == 200)
	{
		HandleMapping(Result->Response->GetContentAsString());
		return true;
	}

	return false;
}

void RepoSrcAssetImporter::HandleMapping(const FString& string)
{
	if (LocalToActorSubmeshMap.Num())
	{
		UE_LOG(Log3DRepo, Error, TEXT("Non-empty Local Map"));
	}

	if (LocalTextureMaps.Num())
	{
		UE_LOG(Log3DRepo, Error, TEXT("Non-empty Local Texture Map"));
	}

	auto &ActorIdMap = actor->IdMap;

	mappings = MakeShareable(new FJsonObject());
	TSharedRef<TJsonReader<TCHAR>> reader = TJsonReaderFactory<TCHAR>::Create(string);
	FJsonSerializer::Deserialize(reader, mappings);

	numSubmeshes = mappings->GetIntegerField(TEXT("numberOfIDs"));

	if (mappings->HasField(TEXT("textures"))) // Textures are completely optional
	{
		auto textures = mappings->GetArrayField(TEXT("textures"));
		for (int32 i = 0; i < textures.Num(); i++)
		{
			auto texture = textures[i]->AsObject();
			FTextureMapInfo map;
			map.Id = FName(*texture->GetStringField(TEXT("id")));
			map.Type = FName(*texture->GetStringField(TEXT("mapType")));
			map.Name = texture->GetStringField(TEXT("name"));
			map.Model = ModelId;
			map.Teamspace = Teamspace;
			LocalTextureMaps.Add(map);
		}
	}

	auto maps = mappings->GetArrayField(TEXT("mapping"));
	for (int32 i = 0; i < maps.Num(); i++)
	{
		auto mapping = maps[i]->AsObject();
		auto name = mapping->GetStringField(TEXT("Name"));

		auto globalIndex = ActorIdMap.AddUnique(name);
		LocalToActorSubmeshMap.Add(globalIndex);

		// Parse the bounds in API coords
		FBox SubMeshBounds;
		FString minString = mapping->GetStringField("min");
		TArray<FString> minStringArr;
		minString.ParseIntoArray(minStringArr, TEXT(" "));
		for (int32 j = 0; j < minStringArr.Num(); j++)
		{
			SubMeshBounds.Min[j] = FCString::Atoi(*minStringArr[j]);
		}
		FString maxString = mapping->GetStringField("max");
		TArray<FString> maxStringArr;
		maxString.ParseIntoArray(maxStringArr, TEXT(" "));
		for (int32 j = 0; j < maxStringArr.Num(); j++)
		{
			SubMeshBounds.Max[j] = FCString::Atoi(*maxStringArr[j]);
		}

		FRepoMeshInfo MeshInfo;
		MeshInfo.ModelId = ModelId;
		MeshInfo.Bounds = SubMeshBounds;

		actor->SubMeshIdToMeshInfoMap.Add(name, MeshInfo);
	}

	// Pull the appearance of the sub-mesh 
	auto materials = mappings->GetArrayField(TEXT("appearance"));
	for (auto material : materials)
	{
		auto name = material->AsObject()->GetStringField(TEXT("name"));
		Material m;

		auto materialJson = material->AsObject()->GetObjectField(TEXT("material"));
		
		m.diffuse = ParseJsonColour(materialJson->GetStringField(TEXT("diffuseColor")));

		if (materialJson->HasField(TEXT("specularColor")))
		{
			m.specular = ParseJsonColour(materialJson->GetStringField(TEXT("specularColor")));
		}
		if(materialJson->HasField(TEXT("transparency")))
		{
			FDefaultValueHelper::ParseFloat(materialJson->GetStringField(TEXT("transparency")), m.transparency);
		}

		mappingMaterials.Add(name, m);
	}

	hasTransparency = false;
	for (auto material : mappingMaterials)
	{
		if (material.Value.transparency != 0)
		{
			hasTransparency = true;
		}
	}

	for (int32 i = 0; i < maps.Num(); i++)
	{
		auto mapping = maps[i]->AsObject();
		auto name = mapping->GetStringField(TEXT("Name")); // in this format (.json.mpc) name is the id
		auto appearance = mapping->GetStringField(TEXT("appearance"));
		auto material = mappingMaterials[appearance];

		material.diffuse.A = 1.0 - material.transparency;

		actor->DiffuseMap->SetParameter(name, material.diffuse);
	}

	// Every time the parameters change we update all map components, not only the ones we explicitly know of,
	// to ensure the materials are initialised to sensible values.
	for (auto component : actor->GetComponents())
	{
		auto map = Cast<URepoSupermeshMapComponent>(component);
		if (map) {
			map->UpdateTexture();
		}
	}

	INC_DWORD_STAT_BY(STAT_TotalObjects, maps.Num())
}

RepoJobRef RepoSrcAssetImporter::HandleSrc(const TArray<uint8>& src)
{
	RepoJobRef job = RepoJob::Create(TEXT("Handle Src ") + this->Asset);

	SCOPE_CYCLE_COUNTER(STAT_HandleSRC);

	auto data = (const uint8*)src.GetData();

	auto preamble = (const uint32*)(data);
	auto srcMagicBit = preamble[0];
	auto srcVersion = preamble[1];
	auto jsonByteSize = preamble[2];

	if (srcMagicBit != SRC_MAGIC_BIT_COMPRESSED && srcMagicBit != SRC_MAGIC_BIT_UNCOMPRESSED)
	{
		UE_LOG(Log3DRepo, Error, TEXT("SRC Magic Bit Mismatch. Expected %d or %d but received %d. Import will be aborted."), SRC_MAGIC_BIT_COMPRESSED, SRC_MAGIC_BIT_UNCOMPRESSED, srcMagicBit);
		job->Complete();
		return job;
	}
	if (srcVersion != SRC_VERSION)
	{
		UE_LOG(Log3DRepo, Error, TEXT("SRC Magic Bit Mismatch. Expected %d but received %d. Import will be aborted."), SRC_VERSION, srcVersion);
		job->Complete();
		return job;
	}

	if (!actor.IsValid())
	{
		job->Complete(); // Abandon the import because the underlying Actor has been destroyed
		return job;
	}

	bool isCompressed = false;
	if (srcMagicBit == SRC_MAGIC_BIT_COMPRESSED)
	{
		isCompressed = true;
	}

	buffer = data + SRC_PREAMBLE_BYTES + jsonByteSize;

	if (isCompressed) //zlib buffer compression is enabled
	{
		auto uncompressedSize = ((uint32_t*)buffer)[0];
		buffer += sizeof(uint32_t);
		auto compressedSize = src.Num() - SRC_PREAMBLE_BYTES - jsonByteSize - sizeof(uint32_t);

		//check() compressedSize

		auto uncompressed = FMemory::Malloc(uncompressedSize);

		FCompression::UncompressMemory(
			NAME_Zlib,
			uncompressed,
			uncompressedSize,
			buffer,
			compressedSize);

		buffer = (uint8_t*)uncompressed;

		INC_MEMORY_STAT_BY(STAT_Uncompressed, uncompressedSize)
	}

	const std::string cstr(reinterpret_cast<const char*>(data + SRC_PREAMBLE_BYTES), jsonByteSize);
	auto jsonStr = FString(cstr.c_str());
	header = MakeShareable(new FJsonObject());
	TSharedRef<TJsonReader<TCHAR>> reader = TJsonReaderFactory<TCHAR>::Create(jsonStr);
	FJsonSerializer::Deserialize(reader, header);

	indexViews = header->GetObjectField(TEXT("accessors"))->GetObjectField(TEXT("indexViews"));
	attributeViews = header->GetObjectField(TEXT("accessors"))->GetObjectField(TEXT("attributeViews"));
	bufferViews = header->GetObjectField(TEXT("bufferViews"));
	bufferChunks = header->GetObjectField(TEXT("bufferChunks"));

	// A 3D Repo scene will have multiple meshes, each delivered as a separate SRC. Within the SRC there are multiple meshes
	// with correspond to the split parts of the SRC's scene mesh.

	auto meshes = header->GetObjectField(TEXT("meshes"))->Values;

	UE_LOG(Log3DRepo, Verbose, TEXT("Creating %d Procedural Meshes for %s."), meshes.Num(), *Uri);

	for (auto mesh_field : meshes)
	{
		SCOPE_CYCLE_COUNTER(STAT_GenerateMesh);

		auto object = mesh_field.Value->AsObject();
		auto attributes = object->GetObjectField(TEXT("attributes"));

		TArray<int32> triangles;
		ResolveIndices(object->GetStringField(TEXT("indices")), triangles);


		TArray<FVector> vertices;
		if (attributes->HasField(TEXT("position")))
		{
			ResolveAttribute<FVector3f>(attributes->GetStringField(TEXT("position")), vertices);
		}

		TArray<FVector> normals;
		if (attributes->HasField(TEXT("normal")))
		{
			ResolveAttribute<FVector3f>(attributes->GetStringField(TEXT("normal")), normals);
		}

		TArray<FVector2D> uv0;
		if (attributes->HasField(TEXT("texcoord")))
		{
			ResolveAttribute<FVector2f>(attributes->GetStringField(TEXT("texcoord")), uv0);
		}

		//Ids are indices into the 'mapping' array provided by the counterpart .json.mpc file.

		TArray<float> ids;
		if (attributes->HasField(TEXT("id")))
		{
			ResolveAttribute<float>(attributes->GetStringField(TEXT("id")), ids);
		}

		TransformCoordinateSystem(vertices);
		TransformCoordinateSystem(normals);
		TransformCoordinateSystem(uv0);

		TArray<FLinearColor> vertexColors; // Empty arrays
		TArray<FProcMeshTangent> tangents;
		TArray<FVector2D> uv1;
		TArray<FVector2D> uv2;
		TArray<FVector2D> uv3;

		GenerateSupermeshMapIndices(ids, uv1); // SupermeshMapIndices relative to the Supermesh itself, and the Actor

		auto mesh = actor->AddProceduralMesh();
		mesh->SetRelativeLocation(Offset);
		mesh->CreateMeshSection_LinearColor(0, vertices, triangles, normals, uv0, uv1, uv2, uv3, vertexColors, tangents, true);

		mesh->SetCollisionProfileName(CollisionProfileName);

		actor->ProceduralMeshFaceMaps.Add(mesh);
		GenerateTriangleIdMap(triangles, ids, actor->ProceduralMeshFaceMaps[mesh].FaceToIdIndex);

		UMaterialInterface* materialPrototype = nullptr;

		if (hasTransparency)
		{
			materialPrototype = materialTranslucent;
		}
		else
		{
			materialPrototype = materialOpaque;
		}

		if (materialPrototype) 
		{
			auto material = TWeakObjectPtr<UMaterialInstanceDynamic>(UMaterialInstanceDynamic::Create(materialPrototype, mesh));
			
			for (auto component : actor->GetComponents())
			{
				auto map = Cast<URepoSupermeshMapComponent>(component);
				if (map) {
					map->ApplyTextureToMaterials(material);
				}
			}

			for (auto& map : LocalTextureMaps)
			{
				job->Add(ResolveTextureAsync(map, material));
			}

			mesh->SetMaterial(0, material.Get()); // We just created the material above so we don't need to check it here.
		}

		Bounds += mesh->CalcLocalBounds().TransformBy(mesh->GetComponentTransform()).GetBox();

		INC_DWORD_STAT_BY(STAT_TotalTriangles, triangles.Num() / 3)
		INC_DWORD_STAT_BY(STAT_TotalVertices, vertices.Num())
	}

	if (isCompressed)
	{
		FMemory::Free((void*)buffer);
	}

	buffer = nullptr;

	job->Complete();
	return job;
}

FVector RepoSrcAssetImporter::TransformCoordinateSystem(FVector v)
{
	return Repo3d::TransformCoordinatesRepoToUnreal(v);
}

void RepoSrcAssetImporter::TransformCoordinateSystem(TArray<FVector2D>& uvs) // As we mirror the geometry we must also swap the UVs to compensate
{
	Repo3d::TransformCoordinatesRepoToUnreal(uvs);
}

void RepoSrcAssetImporter::TransformCoordinateSystem(TArray<FVector>& array) // From the Web/API to Unreal
{
	Repo3d::TransformCoordinatesRepoToUnreal(array);
}

void RepoSrcAssetImporter::GenerateSupermeshMapIndices(TArray<float>& ids, TArray<FVector2D>& uvs)
{
	uvs.SetNumUninitialized(ids.Num());
	for (int32 i = 0; i < ids.Num(); i++)
	{
		uvs[i].X = (float)ids[i];
		uvs[i].Y = (float)LocalToActorSubmeshMap[ids[i]];
	}
}

void RepoSrcAssetImporter::GenerateTriangleIdMap(TArray<int>& triangles, TArray<float>& ids, TArray<int>& triangleIdMap)
{
	auto numTriangles = triangles.Num() / 3;
	triangleIdMap.SetNumUninitialized(numTriangles);
	for (int32 i = 0; i < numTriangles; i++)
	{
		auto index0 = triangles[i * 3];
		auto localId = ids[index0];
		auto globalId = LocalToActorSubmeshMap[localId];
		triangleIdMap[i] = globalId;
	}
}

void RepoSrcAssetImporter::ResolveIndices(const FString& viewName, TArray<int32>& array)
{
	auto indexView = indexViews->GetObjectField(viewName);

	auto viewOffset = indexView->GetIntegerField(TEXT("byteOffset"));
	auto viewCount = indexView->GetIntegerField(TEXT("count"));
	auto viewComponentType = indexView->GetIntegerField(TEXT("componentType"));

	auto bufferView = bufferViews->GetObjectField(indexView->GetStringField(TEXT("bufferView")));
	auto chunks = bufferView->GetArrayField(TEXT("chunks"));

	if (chunks.Num() != 1)
	{
		UE_LOG(Log3DRepo, Error, TEXT("Recevied SRC with bufferView having %d Chunks. This version only supports one chunk per bufferView."), chunks.Num());
		return;
	}

	auto bufferChunk = bufferChunks->GetObjectField(chunks[0]->AsString());

	auto chunkOffset = bufferChunk->GetIntegerField(TEXT("byteOffset"));
	auto chunkLength = bufferChunk->GetIntegerField(TEXT("byteLength"));

	if (viewOffset + (sizeof(uint16) * viewCount) != chunkLength)
	{
		UE_LOG(Log3DRepo, Error, TEXT("Buffer chunk length mismatch. Possible corruption."));
		return;
	}

	array.SetNumUninitialized(viewCount);

	auto data = (const uint16*)(buffer + chunkOffset + viewOffset);
	for (int32 i = 0; i < viewCount; i++)
	{
		array[i] = data[i]; // use for loop here because we are casting uint16 to uint32
	}
}

template <typename T, typename D>
void RepoSrcAssetImporter::ResolveAttribute(const FString& viewName, TArray<D>& array)
{
	auto attributeView = attributeViews->GetObjectField(viewName);

	auto viewOffset = attributeView->GetIntegerField(TEXT("byteOffset"));
	auto viewStride = attributeView->GetIntegerField(TEXT("byteStride"));
	auto viewComponentType = attributeView->GetIntegerField(TEXT("componentType"));
	auto viewType = attributeView->GetStringField(TEXT("type"));
	auto viewCount = attributeView->GetIntegerField(TEXT("count"));

	if (viewComponentType != SRC_COMPONENT_SCALAR)
	{
		UE_LOG(Log3DRepo, Error, TEXT("Recevied SRC Attribute with viewComponentType %d. SRC Attributes must be 32 bit floats (%d)."), viewComponentType, SRC_COMPONENT_SCALAR);
		return;
	}

	if (viewStride != sizeof(T))
	{
		UE_LOG(Log3DRepo, Error, TEXT("Attributes SRC byte stride does not match the size of the expected type."));
		return;
	}

	// Assume the offset and scale are identity.

	auto bufferView = bufferViews->GetObjectField(attributeView->GetStringField(TEXT("bufferView")));
	auto chunks = bufferView->GetArrayField(TEXT("chunks"));

	if (chunks.Num() != 1)
	{
		UE_LOG(Log3DRepo, Error, TEXT("Recevied SRC with bufferView having %d Chunks. This version only supports one chunk per bufferView."), chunks.Num());
		return;
	}

	auto bufferChunk = bufferChunks->GetObjectField(chunks[0]->AsString());

	auto chunkOffset = bufferChunk->GetIntegerField(TEXT("byteOffset"));
	auto chunkLength = bufferChunk->GetIntegerField(TEXT("byteLength"));

	if (viewOffset + (viewStride * viewCount) != chunkLength)
	{
		UE_LOG(Log3DRepo, Error, TEXT("Buffer chunk length mismatch. Possible corruption."));
		return;
	}

	auto data = (T*)(buffer + chunkOffset + viewOffset);
	array.SetNumUninitialized(viewCount);

	if (sizeof(T) == sizeof(D))
	{
		memcpy(array.GetData(), data, chunkLength);
	}
	else
	{
		for (size_t i = 0; i < viewCount; i++)
		{
			array[i] = D(data[i]); // This is essentially only used to convert a vector variant to its LWC equivalent, for which there will always exist this initialiser
		}
	}
}

RepoJobRef RepoSrcAssetImporter::ResolveTextureAsync(const FTextureMapInfo Info, TWeakObjectPtr<UMaterialInstanceDynamic> Material)
{
	return TextureCache->GetTextureAsync(
		Info,
		RepoTextureCache::OnGetTexture::CreateLambda(
			[Material](FTextureMap Map) {
				if (!Material.IsValid())
				{
					return;
				}
				if (!IsValid(Map.Texture))
				{
					return;
				}

				// Check if the material has the parameter slot, so we can warn the user as they may have created a material that is less capable than possible
				TArray<FMaterialParameterInfo> parameterInfo;
				TArray<FGuid> parameterIds;
				Material->GetAllTextureParameterInfo(parameterInfo, parameterIds);

				if (!parameterInfo.ContainsByPredicate([Map](const FMaterialParameterInfo& parameter)
					{
						return parameter.Name.IsEqual(Map.Info.Type);	//FName comparisons are case-insensitive by default
					}))
				{
					UE_LOG(LogTemp, Warning, TEXT("A mesh has the texture map %s, but the Material %s does not have a Texture Parameter with the name '%s' to attach it to."), *Map.Info.Name, *Material->GetFName().ToString(), *Map.Info.Type.ToString());
					return;
				}

				Material->SetTextureParameterValue(Map.Info.Type, Map.Texture);
			}
		)
	);
}