/*
 *	Copyright (C) 2021 3D Repo Ltd
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Affero General Public License as
 *	published by the Free Software Foundation, either version 3 of the
 *	License, or (at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Affero General Public License for more details.
 *
 *	You should have received a copy of the GNU Affero General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "RepoTextureCache.h"
#include "Repo3d.h"
#include "RepoWebRequestManager.h"
#include "ImageUtils.h"
#include "IImageWrapper.h"
#include "IImageWrapperModule.h"
#include "Modules/ModuleManager.h"

RepoTextureCache::RepoTextureCache(TSharedRef<RepoWebRequestManager> WebRequestManager)
	:WebRequestManager(WebRequestManager)
{
}

RepoTextureCache::~RepoTextureCache()
{
}

RepoJobRef RepoTextureCache::GetTextureAsync(const FTextureMapInfo Info, OnGetTexture OnComplete)
{
	// This version of the cache is expected only to be used from the main thread (or at least a single, nominated thread)

	if (!Cache.Contains(Info.Id))
	{
		Cache.Add(Info.Id);
	}

	auto& CacheEntry = Cache[Info.Id];

	// Dereference the weak pointer into a UProperty. This will protect the reference (if valid) until the struct goes out of scope.
	FTextureMap Map;
	Map.Info = Info;
	Map.Texture = CacheEntry.Texture.Get();

	// By design, cached textures can be garbage collected, so check if the actual reference is still valid and re-download if not.
	if (IsValid(Map.Texture))
	{
		OnComplete.Execute(Map);
		return RepoJob::CreateCompleted(TEXT("Return Cached Texture")); // If we have cached the texture then complete this call synchronously
	}
	else
	{
		// The cache is invalid or missing. Check if there is an outstanding request to wait for, 
		// or otherwise create a new one.

		if (!Requests.Contains(Info.Id))
		{
			Requests.Add(Info.Id);
			Requests[Info.Id].Running = false;
		}

		auto& Request = Requests[Info.Id];

		Request.Subscribers.AddLambda(
			[Map, OnComplete](UTexture2D* Texture) {
				auto Params = Map; // Local (modifiable) copy of FTextureMap
				Params.Texture = Texture;
				OnComplete.Execute(Params);
			}
		);

		if (!Request.Running)
		{
			Request.Running = true;
			auto Uri = FString::Printf(TEXT("%s/%s/textures/%s"), *Info.Teamspace, *Info.Model, *Info.Id.ToString());
			auto OnResponse = RepoWebRequestDelegate::CreateLambda(
				[this, Info](TSharedPtr<RepoWebResponse> Result) {
					
					UTexture2D* Texture = nullptr;
					if (Result->bWasSuccessful)
					{
						switch (Result->Response->GetResponseCode())
						{
						case 200:
							Texture = ImportBufferAsTexture2D(Result->Response->GetContent());
							break;
						default:
							UE_LOG(Log3DRepo, Error, TEXT("Failed to download texture (%s); Response Code %d."), *(Result->Request->GetURL()), Result->Response->GetResponseCode());
							break;
						}
					}
					else
					{
						UE_LOG(Log3DRepo, Error, TEXT("Failed to download texture (%s); Response Code %d."), *(Result->Request->GetURL()), Result->Response->GetResponseCode());
					}

					if (!IsValid(Texture))
					{
						UE_LOG(Log3DRepo, Error, TEXT("Failed to load texture (%s); Response Code %d."), *(Result->Request->GetURL()), Result->Response->GetResponseCode());
					}
					else
					{
						auto& CacheEntry = Cache[Info.Id];
						CacheEntry.Texture = TWeakObjectPtr<UTexture2D>(Texture);
						UE_LOG(Log3DRepo, VeryVerbose, TEXT("Downloaded texture (%s)."), *(Result->Request->GetURL()), Result->Response->GetResponseCode());
					}

					auto& Request = Requests[Info.Id];
					Request.Subscribers.Broadcast(Texture);
					Request.Subscribers.Clear();
					Request.Running = false;

				});
			Request.WebRequestJob = WebRequestManager->GetRequest(Uri, OnResponse);
		}

		return Request.WebRequestJob.ToSharedRef();
	}
}

UTexture2D* RepoTextureCache::ImportBufferAsTexture2D(const TArray<uint8>& Buffer)
{
	IImageWrapperModule& ImageWrapperModule = FModuleManager::Get().LoadModuleChecked<IImageWrapperModule>(TEXT("ImageWrapper"));
	EImageFormat Format = ImageWrapperModule.DetectImageFormat(Buffer.GetData(), Buffer.GetAllocatedSize());

	UTexture2D* NewTexture = nullptr;

	if (Format != EImageFormat::Invalid)
	{
		TSharedPtr<IImageWrapper> ImageWrapper = ImageWrapperModule.CreateImageWrapper(Format);
		if (ImageWrapper->SetCompressed((void*)Buffer.GetData(), Buffer.GetAllocatedSize()))
		{
			auto Width = ImageWrapper->GetWidth();
			auto Height = ImageWrapper->GetHeight();

			// These properties set the desired output format for the data returned by GetRaw(). Ensure that
			// the formats passed to GetRaw() and used to set the UTexture2D format match.
			int32 BitDepth = 8;
			ERGBFormat RGBFormat = ERGBFormat::BGRA;
			EPixelFormat PixelFormat = PF_B8G8R8A8;

			const TArray<uint8>* UncompressedDataPtr = nullptr;
#if ENGINE_MAJOR_VERSION <= 4 && ENGINE_MINOR_VERSION < 25
			ImageWrapper->GetRaw(RGBFormat, BitDepth, UncompressedDataPtr);
#else
			TArray<uint8> UncompressedData;
			ImageWrapper->GetRaw(RGBFormat, BitDepth, UncompressedData);
			UncompressedDataPtr = &UncompressedData;
#endif                

			NewTexture = UTexture2D::CreateTransient(Width, Height, PixelFormat);
			if (NewTexture)
			{
				uint8* MipData = static_cast<uint8*>(NewTexture->PlatformData->Mips[0].BulkData.Lock(LOCK_READ_WRITE));

				// Bulk data was already allocated for the correct size when we called CreateTransient above
				FMemory::Memcpy(MipData, UncompressedDataPtr->GetData(), NewTexture->PlatformData->Mips[0].BulkData.GetBulkDataSize());

				NewTexture->PlatformData->Mips[0].BulkData.Unlock();

				NewTexture->UpdateResource();
			}
		}
	}

	return NewTexture;
}