/*
 *	Copyright (C) 2020 3D Repo Ltd
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Affero General Public License as
 *	published by the Free Software Foundation, either version 3 of the
 *	License, or (at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Affero General Public License for more details.
 *
 *	You should have received a copy of the GNU Affero General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "RepoWebRequestManager.h"
#include "Repo3d.h"
#include "RepoTypes.h"
#include "RepoWebRequestHelpers.h"
#include "Containers/UnrealString.h"

DECLARE_MEMORY_STAT(TEXT("Downloaded"), STAT_Downloaded, STATGROUP_Repo3D);
DECLARE_DWORD_ACCUMULATOR_STAT(TEXT("Active Requests"), STAT_ActiveRequests, STATGROUP_Repo3D);

RepoWebRequest::RepoWebRequest(FString Uri, RepoWebRequestDelegate Callback)
	:Job(RepoRequestJob::Create(Uri))
{
	this->uri = Uri;
	this->callback = Callback;
}

RepoWebResponse::RepoWebResponse(RepoJobRef CallbackJob)
	: Job(CallbackJob)
{
}

void RepoWebRequestManager::SetApiKey(FString apikey)
{
	ApiKey = apikey;
}

void RepoWebRequestManager::SetHost(FString host) 
{
	State = Status::Configuring;

	if (host.StartsWith(TEXT("http"))) // User has specified the protocol, which will always be http or https
	{
		Host = host;
	}
	else
	{
		Host = FString::Printf(TEXT("https://%s"), *host);
	}

	RepoWebRequestHelpers::GetApplicationVersion(Owner->GetWebRequestManager(), // This is the TSharedRef equivalent of "this"
		RepoWebRequestHelpers::ApplicationVersionDelegate::CreateLambda(
			[this](TSharedRef<RepoWebRequestHelpers::ApplicationVersion> version)
			{
				if (version->Current == FRepo3dModule::Version)
				{
					// Nothing to do
					UpdateState(Status::Authenticated);
				}
				else if (version->Supported.Contains(FRepo3dModule::Version))
				{
					UpdateState(Status::Authenticated);
					Owner->LogWarning("Your version of 3D Repo for Unreal is supported, but there is a new version available.");
				}
				else
				{
					UpdateState(Status::NotSupported);
					Owner->LogError("Your version of 3D Repo for Unreal is unsupported and will not work. Please upgrade your copy of 3D Repo for Unreal.");
				}
			}
	));
}

void RepoWebRequestManager::UpdateState(Status newState)
{
	State = newState;
	if (State >= Status::Authenticated)
	{
		// We have just authenticated; issue any pending requests
		for (auto Request : Requests)
		{
			GetRequest(Request);
		}
		Requests.Reset();
	}
}

void RepoWebRequestManager::GetRequest(RepoWebRequest Request)
{
	switch (State)
	{
	case Status::None:			// Until the server is configured, queue up any requests
	case Status::Configuring:
	case Status::NotSupported:	// The error message should have already been displayed, so we don't do anything
		Requests.Add(Request);  // (Even if this host is unsupported, the user could still reconnect to a supported one...)
		break;
	case Status::Authenticated:
		GetRequestSync(Request);
		break;
	}
}

void RepoWebRequestManager::GetRequestSync(RepoWebRequest RepoRequest)
{
	FHttpModule::Get().SetHttpTimeout(3600); // the default timeout of 160 seconds is not enough for many models.
	auto HttpRequest = FHttpModule::Get().CreateRequest();
	HttpRequest->SetVerb("GET");

	FString postfix;

	if (!ApiKey.IsEmpty())
	{
		postfix = FString::Printf(TEXT("?key=%s"), *ApiKey);
	}

	auto timestamp = FPlatformTime::Seconds();
	
	HttpRequest->SetURL(FString::Printf(TEXT("%s/api/%s%s"), *Host, *(RepoRequest.uri), *postfix));

	HttpRequest->OnRequestProgress().BindLambda(
		[RepoRequest](FHttpRequestPtr Request, int32 BytesSent, int32 BytesReceived)
		{
			// Use the OnRequestProgress rather than OnProcessRequestComplete for the stats so we can get more 
			// frequent updates which give an indication of how fast the download is progressing.
			auto NewBytesRecieved = BytesReceived - RepoRequest.Job->DownloadJob->TotalBytesRecieved;
			RepoRequest.Job->DownloadJob->TotalBytesRecieved = BytesReceived;
			
			INC_MEMORY_STAT_BY(STAT_Downloaded, NewBytesRecieved);

			// Only emit an update if we have useful progress; the ContentLength member of Request will not be useful, 
			// so wait until the Response is initialised, and even when Response is initialised ContentLength may not 
			// be populated if the response is chunked, for example.
			auto Response = Request->GetResponse();
			if (Response.IsValid() && Response->GetContentLength())
			{
				RepoRequest.Job->DownloadJob->Update(Response->GetContentLength(), BytesReceived);
			}
		}
	);

	HttpRequest->OnProcessRequestComplete().BindLambda(
		[RepoRequest, timestamp](FHttpRequestPtr Request, FHttpResponsePtr Response, bool bWasSuccessful) // note that Unreal should support binding delegates directly, but this function appears to be missing https://docs.unrealengine.com/en-US/Programming/UnrealArchitecture/Delegates/index.html
		{
			DEC_DWORD_STAT_BY(STAT_ActiveRequests, 1);
			RepoRequest.Job->DownloadJob->Complete(Response->GetURL(), Response->GetResponseCode());

			auto result = MakeShared<RepoWebResponse>(RepoRequest.Job->CallbackJob);
			result->bWasSuccessful = bWasSuccessful;
			result->Request = Request;
			result->Response = Response;
			result->Time = FPlatformTime::Seconds() - timestamp;

			RepoRequest.callback.ExecuteIfBound(result);

			RepoRequest.Job->CallbackJob->Complete();
		});

	INC_DWORD_STAT_BY(STAT_ActiveRequests, 1);
	RepoRequest.Job->SetUri(HttpRequest->GetURL());
	HttpRequest->ProcessRequest();
}

RepoJobRef RepoWebRequestManager::GetRequest(FString uri, RepoWebRequestDelegate callback)
{
	RepoWebRequest Request(uri, callback);
	GetRequest(Request);
	return Request.Job;
}

FString RepoWebRequestManager::MakeURI(FString teamspace, FString model, FString revision, FString asset)
{
	if (revision == "")
	{
		revision = TEXT("master/head");
	}
	return (FString::Printf(TEXT("%s/%s/revision/%s/%s"), *teamspace, *model, *revision, *asset));
}