/*
 *	Copyright (C) 2020 3D Repo Ltd
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Affero General Public License as
 *	published by the Free Software Foundation, either version 3 of the
 *	License, or (at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Affero General Public License for more details.
 *
 *	You should have received a copy of the GNU Affero General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include "CoreMinimal.h"
#include "Modules/ModuleManager.h"
#include "Interfaces/IPluginManager.h"
#include "Templates/SharedPointer.h"
#include "RepoTypes.h"
#include "RepoWebRequestManager.h"
#include "RepoSupermeshActor.h"
#include "RepoSupermeshMapComponent.h"
#include "RepoJobs.h"

DECLARE_STATS_GROUP(TEXT("3D Repo Plugin"), STATGROUP_Repo3D, STATCAT_Advanced)

/*
* Declares the log category Log3DRepo which is used throughput the plugin. Most messages are
* Verbose or VeryVerbose and so will not show up by default in the Console log.
* This can be changed using the console with Log [Category] [Verbosity]. For example,
*	Log Log3DRepo Verbose
* Sets the verbosity to Verbose.
*/
DECLARE_LOG_CATEGORY_EXTERN(Log3DRepo, Log, All);

class REPO3D_API Repo3d
{
private:
	FString package;
	UMaterialInterface* opaqueMaterial;
	UMaterialInterface* translucentMaterial;
	TSharedRef<RepoWebRequestManager> manager;
	TSharedRef<class RepoTextureCache> textureCache;

	UMaterialInterface* LoadMaterial(FString fullyQualifiedMatName);

	void FindMaterials();

public:
	Repo3d();

	TSharedRef<RepoWebRequestManager> GetWebRequestManager();
	TSharedRef<class RepoTextureCache> GetTextureCache();

	void SetHost(FString host);
	void SetApiKey(FString key);

	void SetAssetPackage(FString package);

	void SetOpaqueMaterial(FString fullyQualifiedMatName);
	void SetTranslucentMaterial(FString fullyQualifiedMatName);

	RepoLoadModelJobRef LoadModel(FLoadModel params);
	RepoLoadModelJobRef LoadModel(FString teamspace, FString model, FString revision, TWeakObjectPtr<ARepoSupermeshActor> actor);

	DECLARE_MULTICAST_DELEGATE_TwoParams(OnLoadModelDelegate, FLoadModel, FModelInfo);

	/// <summary>
	/// Emitted when a model asset begins loading.
	/// </summary>
	OnLoadModelDelegate OnLoadModel;

	static FVector TransformCoordinatesRepoToUnreal(FVector v);
	static void TransformCoordinatesRepoToUnreal(TArray<FVector>& array);
	static void TransformCoordinatesRepoToUnreal(TArray<FVector2D>& uvs);

	// These log warnings to the user, as well as the UE_LOG
	void LogWarning(FString warning);
	void LogError(FString error);
};


class REPO3D_API FRepo3dModule : public IModuleInterface
{
	static TSharedPtr<Repo3d> singleton;

public:
	static TSharedRef<Repo3d> Get();

	// The Version will be set in StartupModule
	static Version Version;

	/** IModuleInterface implementation */
	virtual void StartupModule() override;
	virtual void ShutdownModule() override;
	virtual bool IsGameModule() const override;
};
