/*
 *	Copyright (C) 2020 3D Repo Ltd
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Affero General Public License as
 *	published by the Free Software Foundation, either version 3 of the
 *	License, or (at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Affero General Public License for more details.
 *
 *	You should have received a copy of the GNU Affero General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include "CoreMinimal.h"
#include "RepoTypes.h"

class RepoJob;
class RepoDownloadJob;
class RepoRequestJob;
class RepoLoadModelJob;

typedef TSharedRef<RepoJob, ESPMode::ThreadSafe> RepoJobRef;
typedef TSharedPtr<RepoJob, ESPMode::ThreadSafe> RepoJobPtr;
typedef TWeakPtr<RepoJob, ESPMode::ThreadSafe> RepoJobWeakPtr;
typedef TSharedRef<RepoDownloadJob, ESPMode::ThreadSafe> RepoDownloadJobRef;
typedef TSharedRef<RepoRequestJob, ESPMode::ThreadSafe> RepoRequestJobRef;
typedef TSharedRef<RepoLoadModelJob, ESPMode::ThreadSafe> RepoLoadModelJobRef;


DECLARE_MULTICAST_DELEGATE_OneParam(RepoJobProgressDelegate, RepoJobPtr);

/**
 * The RepoJob and its subclasses keep track of long running tasks across multiple frames.
 * 
 * Jobs can be used to track progress between different methods and callbacks.
 * Methods should create Jobs when starting tasks that will take some time and for which
 * users should receive feedback on.
 * Jobs are informative only and cannot be used to control the lifetime of the task.
 * 
 * Jobs are hierarchical: Jobs can be added under eachother so higher level jobs
 * automatically track the progress of their component tasks.
 * 
 * For example, an HTTP Request is one Job, and consists of a sub-Job to download the data,
 * and a sub-Job to operate on the downloaded data. The Job that operates on the downloaded
 * data may itself have a number of sub-jobs, each represented as a hierarchy of child
 * Job instances.
 * Only when all sub-jobs in all branches are complete, is the main request considered
 * complete.
 * 
 * Methods can call Update on a Job as often as they like, and the entire hierarhcy
 * will be updated.
 * Classes must call Complete when the job is finished. Calling Complete will automatically
 * set progress to 1, but setting progress to 1 is *not* the same as calling Complete.
 * Update does not need to be called during the lifetime of the Job.
 * 
 * Once a Job is marked as complete, it may never be unmarked, nor may any child jobs
 * be added to it. 
 */
class REPO3D_API RepoJob
{
	/// <summary>
	/// The state of this Job, excluding any child jobs
	/// </summary>
	bool IsCompleted;

	/// <summary>
	/// The progress of this Job, excluding any child jobs
	/// </summary>
	float Progress;
	
	/// <summary>
	/// The total number of Jobs below this object in the whole graph
	/// </summary>
	int TotalJobs;

	/// <summary>
	/// The number of Jobs running below this Job in the whole graph
	/// </summary>
	int CompletedJobs;

	/// <summary>
	/// The overall progress of this job, considering the progress of child jobs
	/// </summary>
	float OverallProgress;

	TArray<RepoJobWeakPtr> Parents;
	TArray<RepoJobRef> Children;

	void Update();

public:
	RepoJob(FName Name);
	~RepoJob();

	/// <summary>
	/// Contains a weak reference to this object. This is used to build the Parents array of other objects.
	/// </summary>
	RepoJobWeakPtr This;

	/// <summary>
	/// User friendly name of the Job
	/// </summary>
	FName Name;

	void Add(RepoJobRef SubJob);
	void Update(float progress);
	void Complete();

	/// <summary>
	/// Whether the entire Job hierarchy is Complete (this job, and all child jobs)
	/// </summary>
	/// <returns></returns>
	bool IsComplete();

	/// <summary>
	/// Returns the number of the jobs that are or were running
	/// </summary>
	int GetNumJobsTotal()
	{
		return TotalJobs;
	}

	/// <summary>
	/// Returns the number of jobs that are currently running
	/// </summary>
	int GetNumJobsCompleted()
	{
		return CompletedJobs;
	}

	/// <summary>
	/// Returns the percentage complete of the job (considering subjobs) between 0 and 1
	/// </summary>
	float GetProgress()
	{
		return OverallProgress;
	}

	RepoJobProgressDelegate OnProgressUpdate;

	static RepoJobRef Create(FName Name);
	static RepoJobRef Create(FString Name);

	/// <summary>
	/// Create an already completed job. This may be used to signal on return that the job already completed during the call.
	/// </summary>
	/// <returns></returns>
	static RepoJobRef CreateCompleted(FString Name);
};

class REPO3D_API RepoLoadModelJob : public RepoJob
{
public:
	RepoLoadModelJob(FLoadModel& Params);

	FLoadModel Params;

	static RepoLoadModelJobRef Create(FLoadModel& Params);
};

class REPO3D_API RepoDownloadJob : public RepoJob
{
public:
	RepoDownloadJob(FName Name);

	int32 TotalBytesRecieved;

	void Update(int32 ContentLength, int32 BytesReceived);
	void Complete(const FString& Uri, int32 ResponseCode);

	static RepoDownloadJobRef Create(FString Name);
};

/*
 * Job representing an HTTP request. This Job consists of two subjobs - download the
 * response and issue the callback provided to the WebRequestManager.
 */
class REPO3D_API RepoRequestJob : public RepoJob
{
public:
	RepoRequestJob(FString Uri);

	RepoDownloadJobRef DownloadJob;
	RepoJobRef CallbackJob;

	// Set the absolute Uri for logging purposes
	void SetUri(FString Uri);

	static RepoRequestJobRef Create(FString Uri);
};