/*
 *	Copyright (C) 2020 3D Repo Ltd
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Affero General Public License as
 *	published by the Free Software Foundation, either version 3 of the
 *	License, or (at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Affero General Public License for more details.
 *
 *	You should have received a copy of the GNU Affero General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include "CoreMinimal.h"
#include "GameFramework/Actor.h"
#include "Math/UnitConversion.h"
#include "RepoStaticSupermeshActor.h"
#include "RepoInterfaces.h"
#include "RepoTypes.h"
#include "RepoSupermeshMapComponent.h"
#include "RepoSupermeshActor.generated.h"

class IAssetTools; // Forward declaration for the static conversion methods. This is not used at runtime.

// (This class is needed because TMap does not support nested containers.)
USTRUCT()
struct REPO3D_API FaceMap
{
	GENERATED_BODY()

	UPROPERTY()
	TArray<int> FaceToIdIndex;
};

/**
 * Top level actor that will host the model/federation data imported by the URepoSrcImporter.
 *
 * At runtime ProdecuralMeshComponents are created and added to this actor, each
 * hosting the supermesh data for the model. Each procedurel mesh's local position be
 * determined by the model offsets.
 *
 * There may be multiple ProdecuralMeshComponents per model, depending on how the
 * model data was batched, but they would all have the same model offset applied to them.
 *
 * Notes:
 * - Federation - A set of models
 * - Supermesh - A single mesh that contains data for a set of objects in the
 * original scene hierarchy. This is handled as such for rendering efficiency. 
 * - Submesh - A portion of the supermesh that represents an object in the
 * original scene hierarchy.
 * - SubmeshId - UUID for the submesh. This is required to get data regarding
 * the object represented by the submesh. 
 *
 * Example 3drepo scene:
 * Model Federation
 *    |
 *    --> Model 1
 *    |     |
 *    |     --> Object 1
 *    |     |
 *    |     --> Object 2
 *    |
 *    --> Model 2
 *          |
 *          --> Object 3
 *          |
 *          --> Object 4
 *          
 * Resulting actor layout:
 * ARepoSupermeshActor
 *    |
 *    --> ProdecuralMeshComponent for Model 1 -|
 *    |                                        | Will contain Object1/2. Local
 *    |                                        | position determined by Model 1 offset.
 *    --> ProdecuralMeshComponent for Model 1 -|
 *    |
 *    --> ProdecuralMeshComponent for Model 2 -|
 *    |										   | Will contain Object3/4. Local
 *    |										   | position determined by Model 2 offset.
 *    --> ProdecuralMeshComponent for Model 2 -|
 *
 */

UCLASS()
class REPO3D_API ARepoSupermeshActor : public AActor, public IRepoTraceable
{
	GENERATED_BODY()

	UPROPERTY(VisibleAnywhere, Category = "Transform Component")
	USceneComponent* Transform;

public:
	UPROPERTY(VisibleAnywhere, Category = "3DRepo Model Data")
	EUnit Units;

	/**
	 * A SupermeshMap Component that stores the diffuse colours of each subobject.
	 * To add additional parameters, subclass ARepoSupermeshActor.
	 */
	UPROPERTY(VisibleAnywhere, Category = "3DRepo Supermeshing Data")
	URepoSupermeshMapComponent* DiffuseMap;

	/**
	 * Contains the Id's of the submeshes/objects within this actor. The UV1.y
	 * attribute of the supermesh geometries contains indices into this array.
	 */
	UPROPERTY(AdvancedDisplay)
	TArray<FString> IdMap;

	/**
	 * Contains the information about each submesh in this Actor, accessible
	 * by their submesh Id
	 */
	UPROPERTY(AdvancedDisplay)
	TMap<FString, FRepoMeshInfo> SubMeshIdToMeshInfoMap;

	/**
	 * Contains all the models loaded under this Actor, accessible by their model Id
	 */
	UPROPERTY(AdvancedDisplay)
	TMap<FString, FModelInfo> ModelIdToModelInfoMap;

	// Sets default values for this actor's properties
	ARepoSupermeshActor();
	
	UProceduralMeshComponent* AddProceduralMesh();
	
	FString GetMeshId(int32 meshIndex);
	
	FRepoMeshInfo GetSubMeshInfo(FString SubMeshId);
	
	/**
	 *	Get the bounds of a given submesh within the model. The bounds are in
	 *	world space in cm.
	 *	@param	SubMeshId		UUID for the submesh
	 */
	FBox GetSubMeshBoundsWorld(FString SubMeshId);
	
	TArray<FString>& GetSubmeshMap();

#if WITH_EDITOR
	ARepoStaticSupermeshActor* AddStaticSupermeshActor();

	// This base actor includes functionality to convert its transient Unreal types into their static equivalents.
	void ConvertToStaticHierarchy();

	// Converts the transient texture Texture into an asset and saves it into the Package provided. Returns the static instance
	// that can be directly applied to a material.
	UTexture2D* ConvertToStaticTexture(UTexture2D* Texture, IAssetTools& AssetTools, UPackage* Package);
#endif

	virtual FString GetMeshIdFromFaceIndex(TWeakObjectPtr<class UPrimitiveComponent> Component, int32 FaceIndex) override;
	virtual TWeakObjectPtr<ARepoSupermeshActor> GetActor();

	// Sets the Parameter on all Dynamic Materials under this ARepoSupermeshActor. This will operate on both static
	// and transient types (before and after the conversion to a static hierarchy).
	void SetMaterialsTextureParameter(FName ParameterName, UTexture* Value);

	// When the hierarchy is created at runtime, this map is used to dereference the Face Indices returned by
	// an FHitResult (which will return ARepoSupermeshActor as the Actor).
	UPROPERTY()
	TMap<UPrimitiveComponent*, FaceMap> ProceduralMeshFaceMaps;

protected:
	// Called when the game starts or when spawned
	virtual void BeginPlay() override;

public:	
	// Called every frame
	virtual void Tick(float DeltaTime) override;

};
