/*
 *	Copyright (C) 2021 3D Repo Ltd
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Affero General Public License as
 *	published by the Free Software Foundation, either version 3 of the
 *	License, or (at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Affero General Public License for more details.
 *
 *	You should have received a copy of the GNU Affero General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include "CoreMinimal.h"
#include "UObject/Class.h"
#include "Engine/Texture.h"
#include "RepoJobs.h"
#include "RepoTextureCache.generated.h"

 /// <summary>
 /// Contains the metadata for a texture map belonging to a model, including everything necessary to retrieve it from the web
 /// </summary>
USTRUCT()
struct FTextureMapInfo
{
	GENERATED_BODY()

	/// <summary>
	/// The usage, or semantic, of the map (e.g. Diffuse or Normal). It is case insensitive by virtue of being an FName.
	/// </summary>
	UPROPERTY()
	FName Type;

	/// <summary>
	/// The UUID of the texture, as it is referred to in the (Repo web API) materials.
	/// </summary>
	UPROPERTY()
	FName Id;

	UPROPERTY()
	FString Teamspace;

	UPROPERTY()
	FString Model;

	UPROPERTY()
	FString Name;
};

/// <summary>
/// Helper struct for providing a native texture instance along with its metadata
/// </summary>
USTRUCT()
struct FTextureMap
{
	GENERATED_BODY()

	/// <summary>
	/// The properties provided when the texture was requested.
	/// </summary>
	UPROPERTY()
	FTextureMapInfo Info;

	/// <summary>
	/// A reference to Texture. This reference will remain valid so long as this struct is in scope, 
	/// but the texture is not guaranteed to be protected simply by being in the cache. Callers must 
	/// assign this elsewhere before returning to prevent it being released.
	/// Texture will not be valid if the texture failed to load, for example if it could not be found.
	/// Use IsValid to check the texture before use. RepoTextureCache will report an error to the log
	/// in the case that the texure fails to load.
	/// </summary>
	UPROPERTY()
	UTexture2D* Texture;
};

/// <summary>
/// Supports loading textures from the Repo GetTexture API into Unreal UTexture2D instances. 
/// All methods should operate through this to avoid re-loading textures unnecessarily.
/// </summary>
/// <remarks>
/// Textures loaded by RepoTextureCache are not protected by design. If a texture ceases to be referenced by a scene object, 
/// it will be released and RepoTextureCache will need to re-download it next time is requested.
/// </remarks>
class REPO3D_API RepoTextureCache
{
public:
	RepoTextureCache(TSharedRef<class RepoWebRequestManager> WebRequestManager);
	~RepoTextureCache();

	DECLARE_DELEGATE_OneParam(OnGetTexture, FTextureMap);

	/// <summary>
	/// Loads a texture asynchronously and calls OnComplete when ready. OnComplete will always be called, even if loading fails,
	/// so callers must check the texture reference in FTextureMap
	/// </summary>
	RepoJobRef GetTextureAsync(FTextureMapInfo Map, OnGetTexture OnComplete);

private:
	TSharedRef<class RepoWebRequestManager> WebRequestManager;

	struct TextureCacheEntry
	{
		TWeakObjectPtr<UTexture2D> Texture;
	};

	// Texture Id's are GUIDs, so are sufficiently unique even without the Teamspace/Model
	// The TWeakObjPtr allows the UTexture to be garbage collected, if all the models using the texture are subsequently deleted.
	TMap<FName, TextureCacheEntry> Cache;

	DECLARE_MULTICAST_DELEGATE_OneParam(OnWebRequestResolved, UTexture2D*);

	struct TextureCacheRequest
	{
		OnWebRequestResolved Subscribers;
		bool Running;
		RepoJobPtr WebRequestJob;
	};

	TMap<FName, TextureCacheRequest> Requests;

	// Creates a Transient Texture2D from (optionally compressed) Image data.
	// (This is identical to FImageUtils::ImportBufferAsTexture2D but it imports all
	// textures with the same bit depth, as the FImageUtils implementation has difficulty
	// with formatting 16 bit images.)
	static UTexture2D* ImportBufferAsTexture2D(const TArray<uint8>& Buffer);
};