/*
 *	Copyright (C) 2020 3D Repo Ltd
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Affero General Public License as
 *	published by the Free Software Foundation, either version 3 of the
 *	License, or (at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Affero General Public License for more details.
 *
 *	You should have received a copy of the GNU Affero General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include "CoreMinimal.h"
#include "RepoTypes.generated.h"


/**
 * Holds information regarding an imported 3drepo model.
 */
USTRUCT()
struct FModelInfo
{
	GENERATED_BODY()

	FString ModelId;
	FString Teamspace;
	FVector Offset; // In the API coordinate system
};

/**
 * Holds information relating to a submesh within a supermesh
 */
USTRUCT()
struct FRepoMeshInfo
{
	GENERATED_BODY()

	FString ModelId; // Id of the model to which this submesh belongs
	FBox Bounds; // In the API coordinate system
};

/**
 * Version of the plugin
 */
class Version
{
public:
	int Major;
	int Minor;
	int Revision;

	Version();

	static Version Parse(const FString& version);
	friend bool operator==(const Version& lhs, const Version& rhs)
	{
		return FMemory::Memcmp(&lhs, &rhs, sizeof(Version)) == 0;
	}

	FString ToString()
	{
		return FString::Printf(TEXT("%d.%d.%d"), Major, Minor, Revision);
	}
};

USTRUCT()
struct FLoadModel
{
	GENERATED_BODY()

	/// <summary>
	/// The Teamspace/Database that contains the model. This can be found from the Teamspaces view on the Web.
	/// </summary>
	FString Teamspace;
	
	/// <summary>
	/// The UUID of the Model or Federation to load. This can be found from the Settings dialog of the Model or Federation on the Web.
	/// </summary>
	FString Model;

	/// <summary>
	/// Optional. The Revision of the model to load. This can be a tag or a UUID.
	/// </summary>
	FString Revision;

	/// <summary>
	/// A pointer to the Actor under which to load the model. An Actor can be passed in multiple times to have multiple models loaded under it.
	/// </summary>
	TWeakObjectPtr<class ARepoSupermeshActor> Actor;

	/// <summary>
	/// Optional. The Name of the Collision Profile to apply to the loaded geometry. If not set this will default to IgnoreOnlyPawn. 
	/// Collision Profiles can be configured in Project Settings -> Engine -> Collision.
	/// </summary>
	FName CollisionProfileName;
};