/*
 *	Copyright (C) 2020 3D Repo Ltd
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Affero General Public License as
 *	published by the Free Software Foundation, either version 3 of the
 *	License, or (at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Affero General Public License for more details.
 *
 *	You should have received a copy of the GNU Affero General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "SampleWebRequestHelper.h"
#include "Repo3d.h"
#include "Json.h"

class RepoWebGetTreeContext
{
public:
	class SubTree
	{
	public:
		FString _id;
		FString url;
	};

	TSharedRef<RepoWebRequestManager> manager;
	TSharedRef<SampleWebRequestHelpers::Tree> tree;
	SampleWebRequestHelpers::GetTreeDelegate callback;

	// Contains the subtrees still to be processed.
	TArray<RepoWebGetTreeContext::SubTree> subTrees;

	RepoWebGetTreeContext(TSharedRef<RepoWebRequestManager> Manager, SampleWebRequestHelpers::GetTreeDelegate Callback)
		:manager(Manager), tree(MakeShared<SampleWebRequestHelpers::Tree>()), callback(Callback)
	{
	}

	TSharedRef<SampleWebRequestHelpers::Tree::Node> SearchTree(TSharedRef<SampleWebRequestHelpers::Tree::Node> node, FString Id)
	{
		if (node->_id == Id)
		{
			return node;
		}

		for (auto child : node->children)
		{
			auto found = SearchTree(child, Id);
			if (found != child)
			{
				return found;
			}
		}

		return node;
	}

	TSharedRef<SampleWebRequestHelpers::Tree::Node> ParseNode(const TSharedPtr<FJsonObject> jsonNode)
	{
		TSharedRef< SampleWebRequestHelpers::Tree::Node> node = MakeShared< SampleWebRequestHelpers::Tree::Node>();

		node->account = jsonNode->GetStringField("account");
		node->project = jsonNode->GetStringField("project");
		node->type = jsonNode->GetStringField("type");
		if (jsonNode->HasField("name"))
		{
			node->name = jsonNode->GetStringField("name");
		}
		node->path = jsonNode->GetStringField("path");
		node->_id = jsonNode->GetStringField("_id");
		node->shared_id = jsonNode->GetStringField("shared_id");
		node->toggleState = jsonNode->GetBoolField("toggleState");

		if (jsonNode->HasField("children")) {
			auto jsonChildren = jsonNode->GetArrayField("children");
			for (auto jsonChild : jsonChildren)
			{
				node->children.Add(ParseNode(jsonChild->AsObject()));
			}
		}

		return node;
	}

	void ParseIdMap(const TSharedPtr<FJsonObject> jsonNode)
	{
		for (auto field : jsonNode->Values)
		{
			tree->idToName.Add(field.Key, field.Value->AsString());
		}
	}

	void ParseTree(const FString& _id, const TSharedPtr<FJsonObject> jsonTree)
	{
		auto node = ParseNode(jsonTree->GetObjectField("nodes"));

		if (tree->nodes->_id == "") // this is an emptry tree (i.e. we have got the root tree)
		{
			tree->nodes = node;
		}
		else // otherwise find the parent node of the subtree and add it under that
		{
			auto parent = RepoWebGetTreeContext::SearchTree(tree->nodes, _id);
			parent->children.Add(node);
		}

		ParseIdMap(jsonTree->GetObjectField("idToName"));
	}

	// Create a static function that takes its enclosing type as a parameter, so that when the lambda closures are 
	// created it is the TSharedRef, now the raw 'this' pointer, that is stored, and so the RepoWebGetTreeContext 
	// will not be prematurely destroyed.
	static void OnSubTreeComplete(TSharedRef<RepoWebGetTreeContext> context)
	{
		if (context->subTrees.Num() > 0)
		{
			auto subTree = context->subTrees.Pop();
			context->manager->GetRequest(subTree.url,
				RepoWebRequestDelegate::CreateLambda(
					[context, subTree](TSharedPtr<RepoWebResponse> Result) {
						if (Result->bWasSuccessful) {
							auto string = Result->Response->GetContentAsString();
							auto reader = TJsonReaderFactory<TCHAR>::Create(string);
							TSharedPtr<FJsonObject> jsonResponse = MakeShareable(new FJsonObject());
							FJsonSerializer::Deserialize(reader, jsonResponse);

							context->ParseTree(subTree._id, jsonResponse->GetObjectField("mainTree"));

							auto jsonSubTrees = jsonResponse->GetArrayField("subTrees");
							for (auto jsonSubTreeValue : jsonSubTrees)
							{
								auto jsonSubTree = jsonSubTreeValue->AsObject();
								RepoWebGetTreeContext::SubTree newSubTree;
								newSubTree.url = jsonSubTree->GetStringField("url");
								newSubTree._id = jsonSubTree->GetStringField("_id");

								context->subTrees.Push(newSubTree);
							}

							// Make a recursive call to this function, that will either start processing the subtrees added above,
							// or issue the callback
							OnSubTreeComplete(context);
						}

					})
			);
		}
		else
		{
			context->callback.ExecuteIfBound(context->tree);
		}
	}
};

void SampleWebRequestHelpers::GetTree(TSharedRef<RepoWebRequestManager> manager, FString teamspace, FString model, FString revision, GetTreeDelegate callback)
{
	RepoWebGetTreeContext::SubTree baseTree;
	baseTree._id = "";
	baseTree.url = RepoWebRequestManager::MakeURI(teamspace, model, revision, "fulltree.json");

	TSharedRef<RepoWebGetTreeContext> context = MakeShared<RepoWebGetTreeContext>(manager, callback);
	context->subTrees.Push(baseTree);

	// The GetTree API includes a field with optional subtrees. This function works by maintaining a list of subtrees and re-entering itself
	// on each callback from the Web Request Manager. The callback then either initates another request for the next outstanding tree (if any)
	// or calls the Callback provided above when there are no more trees, thereby fetching each subtree in order, one after the other.
	RepoWebGetTreeContext::OnSubTreeComplete(context);
}
