/*
 *	Copyright (C) 2020 3D Repo Ltd
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Affero General Public License as
 *	published by the Free Software Foundation, either version 3 of the
 *	License, or (at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Affero General Public License for more details.
 *
 *	You should have received a copy of the GNU Affero General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "repounreal.h"
#include "Modules/ModuleManager.h"
#include "CoreMinimal.h"

/*
* Define the log category Log3DRepoSample declared in repounreal.h.
* Change the verbosity for this, and the plugin's category - Log3DRepo -
* using the console, for example,
*	Log Log3DRepo Verbose
*/

DEFINE_LOG_CATEGORY(Log3DRepoSample);

/*
 * Before beginning, add "Repo3d" to the list of PrivateDependencyModuleNames 
 * in your project's build script. This sample project already has this set up.
 * You can find the bulid script under the project folder <projectName>.Buld.cs
 */ 

#include "Repo3d.h"

/* 
 * This sample will use the sample user Actor subclass. You can subclass 
 * ARepoSupermeshActor to control how 3D Repo assets appear and behave.
 */

#include "SampleRepoSupermeshActor.h"

/*
 * This file has optimisations off, so the debugger can be used to see how the
 * samples function without variables being optimised away.
 */

#pragma optimize("", off)

/*
 * This sample module demonstrates how to use the 3D Repo Plug-in for Unreal. 
 * The module registers an Unreal Console Command that can be invoked at both 
 * design- and run-time by bringing up the console (tilde key). The command 
 * executes the ImportModel() below. This method logs into 3D Repo and loads 
 * a model into the scene.
 * 
 * Subclass from IModuleInterface to register as an engine module, so we get 
 * the StartupModule() callback. If you are building a UI, you may want to 
 * subclass from the default game implementation
 */

class SampleGameModule : public IModuleInterface 
{
public:
	SampleGameModule()
	{
		UE_LOG(Log3DRepoSample, Log, TEXT("Loaded 3D Repo Sample Game Module."));
	}

	// The URL of the 3drepo instance to get the models from
	FString url = TEXT("www.3drepo.io");

	// The Unreal plugin uses API keys for authentication. Generate a key from your user profile
	// in the 3drepo instance you are using
	FString apiKey = TEXT("7d0573c6142a95477f9fb7160a10c721");

	// Called by the 3drepoimportdemomodel command
	void ImportDemoModel(UWorld* world)
	{
		ImportModel("Repo3DDemo", "af42d0e0-76b2-11eb-83fd-51b9ea886eaf", "", world);
	}

	// Called by the 3drepoimport command
	void ImportModel(const TArray<FString>& args, UWorld* world)
	{
		if (!CheckImportModelArgs(args))
		{
			return;
		}

		auto teamspace = args[0];
		auto model = args[1];
		auto revision = FString("");

		if (args.Num() > 2)
		{
			revision = args[2];
		}

		ImportModel(teamspace, model, revision, world);
	}

	/*
	 * ImportModel shows how to use the plugin's API. First, the plugin must have
	 * a host and API key for authentication. The Plugin also needs some parameters 
	 * about how it should interact with Unreal, including in which World to create the 
	 * models, and what materials to use.
	 * 
	 * Once these are set, the command to import the model can be given.
	 */
	void ImportModel(FString Teamspace, FString Model, FString Revision, UWorld* World)
	{
		// Get an interface to the 3D Repo Plugin, and login (or set your API key).

		// The Get() method is a convenience function that returns a singleton 
		// (so one instance remains logged in application wide). You can also create
		// a new instance of Repo3d.
		auto Repo3d = FRepo3dModule::Get();

		Repo3d->SetHost(url);
		Repo3d->SetApiKey(apiKey);

		// Prepare the import by indicating where to find the supermesh-aware 
		// materials. 
		// This sample project comes with two materials to build on. 
		// If no material is specified, the Unreal default will be applied.
		Repo3d->SetAssetPackage("Game");
		Repo3d->SetOpaqueMaterial("RepoHighlightMaterial");
		Repo3d->SetTranslucentMaterial("RepoHighlightMaterialTranslucent");

		// Provide an Actor in the world to import the model into. This Actor
		// will store the model & revision of the last imported model and have its 
		// scale set to match the model units, so it is recommended to create a new 
		// one for each import. 
		// This sample gets the World to create the Actor in from the Console 
		// Command Callback.
		// While the plugin is using a weak pointer to a ARepoSupermeshActor, note
		// the type actually instantiated is the ASampleRepoSupermeshActor subclass.
		// Use Cast<> or IsA<> to work out the true type of the Actors in callbacks.
		auto SampleActor = TWeakObjectPtr<ASampleRepoSupermeshActor>((ASampleRepoSupermeshActor*)World->SpawnActor(ASampleRepoSupermeshActor::StaticClass()));

		// Finally, specify the teamspace, model & revision to load.
		auto Job = Repo3d->LoadModel(Teamspace, Model, Revision, SampleActor);
		SampleActor->Teamspace = Teamspace;
		SampleActor->Model = Model;
		SampleActor->Revision = Revision;

		// The Job returned by the LoadModel() method can be used to monitor the
		// loading progress. The IsComplete member can be used to determine whether
		// the entire model has finished importing.
		// Turn on the Repo3D Stats, e.g. 'STAT REPO3D' in the console, to see more 
		// detailed statistics such as the amount of data downloaded, and the number 
		// of objects created.
		Job->OnProgressUpdate.AddLambda(
			[](RepoJobPtr job)
			{
				UE_LOG(Log3DRepoSample, Log, TEXT("Loading... %d/%d Jobs"), job->GetNumJobsCompleted(), job->GetNumJobsTotal());
				if (job->IsComplete())
				{
					UE_LOG(Log3DRepoSample, Log, TEXT("Loading Complete."));
				}
			}
		);
	}

	/*
	 * By default, models are imported as transient assets, both at Design and 
	 * Run-time. This reduces loading time and overhead, but prevents the scene
	 * being serialised to a map or PAK.
	 * When running in the Editor, it is possible to convert an imported Model 
	 * into a hierarchy of StaticMeshes and StaticMeshComponents, that can be 
	 * treated like any other Unreal Asset.
	 * The function below converts the ARepoSupermeshActors in the current world
	 * into a form that can be saved by the Editor.
	 */
#if WITH_EDITOR
	void Convert(UWorld* world)
	{
		TArray<ARepoSupermeshActor*> RepoActors;
		for (auto level : world->GetLevels())
		{
			for (auto actor : level->Actors)
			{
				auto repoActor = Cast<ARepoSupermeshActor>(actor);
				if (repoActor && !repoActor->IsActorBeingDestroyed())
				{
					RepoActors.Add(repoActor);
				}
			}
		}

		for (auto repoActor : RepoActors)
		{
			repoActor->ConvertToStaticHierarchy();
		}
	}
#endif

	// This is called by the 3drepowebconfig command
	void Configure3dRepoInstance(const TArray<FString>& args)
	{
		if (args.Num() > 1)
		{
			url = args[0];
		}
		
		if (args.Num() > 0)
		{
			apiKey = args[1];
		}

		auto repo3d = FRepo3dModule::Get();
		repo3d->SetHost(url);
		repo3d->SetApiKey(apiKey);

		UE_LOG(Log3DRepoSample, Log, TEXT("Web settings succesfully configured"));
	}

	virtual void StartupModule() override
	{
		// Registers the ImportModel() callback under the command 3drepoimport
		IConsoleManager::Get().RegisterConsoleCommand(
			TEXT("3drepoimport"),
			TEXT("Import a file from 3dRepo.io (arguments: teamspace modelid [revisiontag])"),
			FConsoleCommandWithWorldAndArgsDelegate::CreateRaw(this, &SampleGameModule::ImportModel),
			0);
		IConsoleManager::Get().RegisterConsoleCommand(
			TEXT("3drepoimportdemomodel"),
			TEXT("Import the demonstration model from 3dRepo.io"),
			FConsoleCommandWithWorldDelegate::CreateRaw(this, &SampleGameModule::ImportDemoModel),
			0);
		IConsoleManager::Get().RegisterConsoleCommand(
			TEXT("3drepowebconfig"),
			TEXT("Configure the instance of 3drepo to use (arguments: url api_key"),
			FConsoleCommandWithArgsDelegate::CreateRaw(this, &SampleGameModule::Configure3dRepoInstance),
			0);
#if WITH_EDITOR // The conversion to static types can only happen in the Editor
		IConsoleManager::Get().RegisterConsoleCommand(
			TEXT("3drepoconvert"),
			TEXT("Convert scene to use Unreal static classes (for further inspection in the Editor, or for saving to a PAK)"),
			FConsoleCommandWithWorldDelegate::CreateRaw(this, &SampleGameModule::Convert),
			0);
#endif
	}

private:

	bool CheckImportModelArgs(const TArray<FString>& args)
	{
		if (apiKey == TEXT("") || url == TEXT(""))
		{
			UE_LOG(Log3DRepoSample, Log, TEXT("API key or URL is not configured. Either set it in the code or use the 3drepowebconfig command"));
			return false;
		}

		if (args.Num() < 2)
		{
			UE_LOG(Log3DRepoSample, Log, TEXT("Incorrect number of arguments. Command requires 2 or 3 arguments: teamspace, model id [and revision]"));
			return false;
		}

		return true;
	}
};

#pragma optimize("", on)

IMPLEMENT_PRIMARY_GAME_MODULE(SampleGameModule, repounreal, "repounreal" );