/*
 *	Copyright (C) 2020 3D Repo Ltd
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Affero General Public License as
 *	published by the Free Software Foundation, either version 3 of the
 *	License, or (at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Affero General Public License for more details.
 *
 *	You should have received a copy of the GNU Affero General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "repounrealGameModeBase.h"
#include "repounreal.h"
#include "RepoSupermeshActor.h"
#include "Repo3d.h"
#include "SampleRepoSupermeshActor.h"
#include "SampleWebRequestHelper.h"
#include <RepoInterfaces.h>
#include <Engine/World.h>

ArepounrealGameModeBase::ArepounrealGameModeBase()
{
	this->PlayerControllerClass = ArepounrealPlayerController::StaticClass();
}

#pragma optimize("", off)

ArepounrealPlayerController::ArepounrealPlayerController()
{
	this->bShowMouseCursor = true;
	this->bEnableMouseOverEvents = true;
	this->bEnableClickEvents = true;
}

void ArepounrealPlayerController::BeginPlay()
{
	FInputModeGameAndUI inputMode;
	SetInputMode(inputMode);

	// This code binds methods to Actions. The Actions are configured by name in Project Settings -> Input -> Bindings -> Action Mappings

	InputComponent->BindAction("OnMouseClick", EInputEvent::IE_Pressed, this, &ArepounrealPlayerController::OnMouseClick);
}

void ArepounrealPlayerController::OnMouseClick()
{
	auto mesh = GetMeshIdUnderCursor();

	if (mesh.Valid) 
	{
		// Print the Id to the console

		UE_LOG(Log3DRepoSample, Log, TEXT("Selected Mesh %s"), *(mesh.Id));

		// Use the ASampleRepoSupermeshActor's highlight map to highlight the object by changing it's colour.
		// Create your own subclass of ARepoSupermeshActor to implement other behaviours.
		// (Use the Cast<> function to check if the ARepoSupermeshActor returned by GetActor() is your subclass.)

		auto sampleActor = Cast<ASampleRepoSupermeshActor>(mesh.Traceable->GetActor());
		if (sampleActor)
		{
			auto highlighted = sampleActor->HighlightMap->GetParameter(mesh.Id);
			highlighted.X = 1 - highlighted.X;
			sampleActor->HighlightMap->SetParameter(mesh.Id, highlighted);
		}
		else
		{
			mesh.Traceable->GetActor()->DiffuseMap->SetParameter(mesh.Id, FVector4(1.0, 1.0, 1.0, 1.0));
		}


		// We can also use the Web API's to return information about the object, including BIM data
		// The API reference can be found at https://3drepo.github.io/3drepo.io/
		// Below, we return the object name from the fullTree.json.
		// Parsing the API responses is the responsibility of the client code. The code above uses a
		// sample helper that shows how the GetTree API might be used.

		// This method requests the Tree for every selection. In a real application the tree should
		// be read once when the model is loaded.

		if (FRepo3dModule::Get()->GetWebRequestManager()->GetState() < RepoWebRequestManager::Status::Configuring)
		{
			UE_LOG(Log3DRepoSample, Log, TEXT("The URL and API key have not been specified. If you have re-loaded this project ensure that the 3DRepo plugin is configured. See the 3drepowebconfig command."));
		}
		else
		{
			SampleWebRequestHelpers::GetTree(
				FRepo3dModule::Get()->GetWebRequestManager(),
				sampleActor->Teamspace,
				sampleActor->Model,
				sampleActor->Revision,
				SampleWebRequestHelpers::GetTreeDelegate::CreateLambda(
					[mesh](TSharedRef<SampleWebRequestHelpers::Tree> tree)
					{
						if (tree->idToName.Contains(mesh.Id))
						{
							auto name = tree->idToName[mesh.Id];
							UE_LOG(Log3DRepoSample, Log, TEXT("Selected Mesh Name %s"), *name);
						}
					}
			));
		}
	}
}

MeshIdSelection ArepounrealPlayerController::GetMeshIdUnderCursor()
{
	/* 
	 * This method shows how to get the identity of the object under the mouse.
	 * It uses Unreal's LineTracing API to get the Actor (and Component) and FaceIndex under the cursor.
	 * 
	 * If the Actor is one of the Repo Plugin's Actors, then it will have an interface that can be used 
	 * to find the Mesh ID. 
	 * 
	 * The Mesh ID can then be used with the 3DRepo Web APIs to query information about the mesh,
	 * or change how it appears using flag textures and appropriate Materials.
	 */

	FVector WorldLocation;
	FVector WorldDirection;

	DeprojectMousePositionToWorld(WorldLocation, WorldDirection);

	// Queries are performed against a set of objects in the world. The world is set for a PlayerController instance (which this class is).

	auto world = this->GetWorld();

	// We set up to trace against Complex colliders. This is necessary because we need the individual faces to perform the Mesh Id lookup.

	// (PlayerController has a number of methods such as GetHitResultUnderCursor(), but we need to set up our own FCollisionQueryParams because
	// bReturnFaceIndex is set to false by default.)

	FCollisionQueryParams Query;
	Query.bReturnFaceIndex = true;
	Query.bReturnPhysicalMaterial = false;
	Query.bTraceComplex = true;

	FCollisionResponseParams Response;

	// (It is also possible to use the asynchronous version of this API (UWorld::AsyncLineTraceByChannel))

	FHitResult OutHit;
	world->LineTraceSingleByChannel(
		OutHit,
		WorldLocation,
		WorldLocation + WorldDirection * HitResultTraceDistance, // This distance is that used by GetHitResultUnderCursor and similar.
		ECollisionChannel::ECC_Visibility,
		Query,
		Response
	);

	// Check if we've cliked on a 3DRepo UObject that is capable of supplying a mesh Id by checking for the appropriate interface

	FString meshId;
	auto repoTraceable = Cast<IRepoTraceable>(OutHit.GetActor());
	if (repoTraceable)
	{
		meshId = repoTraceable->GetMeshIdFromFaceIndex(OutHit.Component, OutHit.FaceIndex);
	}

	MeshIdSelection result;
	result.Valid = meshId.Len() > 0;
	result.Id = meshId;
	result.WorldLocation = OutHit.ImpactPoint;

#if ENGINE_MAJOR_VERSION > 4
	result.Traceable = TWeakInterfacePtr<IRepoTraceable>(repoTraceable);
#else
	result.Traceable = TWeakInterfacePtr<IRepoTraceable>(*repoTraceable);
#endif

	return result;
}

#pragma optimize("", on)

